--[[
    Utilities to handle the module declaration files.
]]

-- Check if the type of the argument is 'table'
function ___checkTable(tbl)
    if type(tbl) ~= "table" then
        error("The passed argument is of type " .. type(tbl) .. " and not table", 3)
    end
end

-- Local private function: try to import options from a table into a job or a
-- module declaration.
function ___tryImportOptions(thing, tbl)
    ___checkTable(tbl)
    if tbl["options"] then
        ___checkTable(tbl.options)
        thing:setOptions(tbl.options)
    end
end

-- The option declaration block:
--      DeclareOption { opt1 = val1, opt2 = val2 }
function DeclareOption(opt)
    ___checkTable(opt)
    local option = OptionDeclaration()
    for o, val in pairs(opt) do
        option:addOptionContent(o, val)
    end
    return option
end

-- The function declaration block
--      DeclareFunction { name }
function DeclareFunction(tbl)
    ___checkTable(tbl)
    if #tbl == 1 and type(tbl[1]) == "string" then
        local func = FunctionDeclaration()
        func:setName(tbl[1])
        return func
    end
    error("invalid number of arguments in the DeclareFuncton block", 2)
end

-- The function import block
--      ImportFunction { script, name }
function ImportFunction(tbl)
    ___checkTable(tbl)
    if #tbl == 2 and type(tbl[1]) == "string" and type(tbl[2]) == "string" then
        local func = FunctionDeclaration()
        func:setImportFromScript(tbl[1])
        func:setName(tbl[2])
        return func
    end
    error("invalid number of arguments in the ImportFunction block", 2)
end

-- The job declaration block
--      DeclareJob { name }
--      DeclareJob { name, options = { ... } }
function DeclareJob(tbl)
    ___checkTable(tbl)
    if #tbl >= 1 and #tbl <= 2 and tbl[1] and type(tbl[1]) == "string" then
        local job = JobDeclaration()
        job:setName(tbl[1])
        ___tryImportOptions(job, tbl)
        return job
    end
    error("invalid DeclareJob block, need at least a name", 2)
end

-- The job import block
--      ImportJob { script, job }
--      ImportJob { script, job, options = { ... } }
function ImportJob(tbl)
    ___checkTable(tbl)
    if #tbl >= 2 and #tbl <= 3 and type(tbl[1]) == "string" and type(tbl[2]) == "string" then
        local job = JobDeclaration()
        job:setName(tbl[2])
        job:setImportFromScript(tbl[1])
        ___tryImportOptions(job, tbl)
        return job
    end
    error("invalid ImportJob block, need at least a script and a name", 2)
end

-- The module declaration block
--      DeclareModule {
--          name = ..., description = ..., author = ...,
--          revision = ..., imports = { ... }, options = { ... },
--          fuctions = { ... }, jobs = { ... }
--      }
--      DeclareModule { name = ..., author = ... }
function DeclareModule(tbl)
    ___checkTable(tbl)
    if not tbl.name   then error("Invalid DeclareModule block, Missing name",   2) end
    if not tbl.author then error("Invalid DeclareModule block, Missing author", 2) end
    local module = ModuleDeclaration()

    -- Standard things
    module:setName(tbl.name)
    module:setAuthorName(tbl.author)

    -- Optional simple things
    if tbl.description then module:setDescription(tbl.description) end
    if tbl.revision    then module:setRevision(tbl.revision) end
    if tbl.imports     then module:setImports(tbl.imports) end

    -- A little more complicated thing
    if tbl.jobs      then module:setJobs(tbl.jobs) end
    if tbl.functions then module:setFunctions(tbl.functions) end

    ___tryImportOptions(module, tbl)
    -- Register jobs and functions in runtime now that they are added to the module!
    for _, job in ipairs(tbl.jobs)      do job:pushToRuntime() end
    for _, fun in ipairs(tbl.functions) do fun:pushToRuntime() end

    -- Register the module!
    module:pushToRuntime()
    return module
end
