#pragma once

#ifndef __cplusplus
#error "This is a C++ header"
#endif

#include "PreCompiledHeaders.hh"
#include "VivyApplication.hh"
#include "Lib/Log.hh"
#include "Lib/Utils.hh"
#include "Lib/Uuid.hh"

#define VIVY_STORAGE_CLASS(theClassName, theDocumentName) \
    VIVY_UNMOVABLE_OBJECT(theClassName)                   \
    friend CRTPStore<theClassName, theDocumentName>;

namespace Vivy
{
// For example:
// VivyDocumentStore final : public
//      CRTPStore<VivyDocumentStore,    <- Child class
//                VivyDocument>         <- Document type to hold
template <class Store, class Document> class CRTPStore {
    VIVY_UNMOVABLE_OBJECT(CRTPStore)

protected:
    VIVY_APP_LOGGABLE_OBJECT(CRTPStore, logger)

    std::map<Uuid, std::shared_ptr<Document>> documents{};
    uint newDocumentNumber{ 1 };
    static inline const QString newDocumentBaseName = QStringLiteral("Untitled ");

protected:
    explicit CRTPStore() noexcept = default;

public:
    // Load a document by its name
    std::shared_ptr<Document> loadDocument(const QString &file)
    {
        auto ret      = std::make_shared<Document>(file);
        const Uuid id = ret->getUuid();

        if (documents.count(id) >= 1)
            throw std::runtime_error("Document already loaded");

        documents[id] = ret;
        return ret;
    }

    // Close a document, please be sure to not used any of the dangling
    // references to the closed document...
    void closeDocument(const Uuid &uuid) noexcept
    {
        auto *self = static_cast<Store *>(this);
        self->documents.erase(uuid);
    }

    // Get stored documents
    std::shared_ptr<Document> getDocument(const Uuid &uuid) const
    {
        auto *self = static_cast<Store *>(this);

        if (!self->documents.contains(uuid)) {
            logError() << "Couldn't find the document " << uuid;
            throw std::runtime_error("Can't find the document");
        }

        return self->documents.value(uuid);
    }

    // Get to see if a document is already present or not
    [[nodiscard("handle-it")]] bool isDocumentPresent(const Uuid &uuid) noexcept
    {
        auto *self = static_cast<Store *>(this);
        return self->documents.count(uuid) >= 1;
    }
};
}
