#include "PreCompiledHeaders.hh"
#include "Lib/Script/CRTPLuaScriptObject.hh"

using namespace Vivy::Script;

bool
ModuleDeclaration::resolvModules(LuaContext *const context) noexcept
{
    for (const auto &str : importNames) {
        const ModuleDeclaration *const mod = context->getModule(str);
        if (mod)
            importedModules.emplace_back(mod);

        else
            context->failedWith("Failed to find needed module: " + str);
    }
    return true;
}

int
ModuleDeclaration::importMethod(lua_State *const L) noexcept
{
    auto *const context = LuaContext::getContext(L);
    luaL_checktype(L, 2, LUA_TTABLE);

    // Get the module name
    lua_pushinteger(L, 1); // Stack is { top: int, table, self }
    lua_gettable(L, 2);    // Stack is { top: str, table, self }
    const std::string_view mod = CHECK_STRING_VIEW(L, 3);

    // Import a module
    context->getModule(mod); // Stack is { top: nil | module, str, table, self }

    // Import a function if needed
    if (lua_rawlen(L, 2) == 2) {
        lua_pushinteger(L, 2); // Stack is { top: int, module, str, table, self }
        lua_gettable(L, 2);    // Stack is { top: str, module, str, table, self }
        const std::string_view func         = CHECK_STRING_VIEW(L, 5);
        FunctionDeclaration *const function = context->getFunction(mod, func);
        if (function)
            function->pushLuaFunction(L);
        // Stack is { top: nil | function, FunctionDeclaration, str, module, str, table, self }
    }

    // Top of stack is nil | function | module
    return 1;
}

int
ModuleDeclaration::exportMethod(lua_State *const L) noexcept
{
    ModuleDeclaration *const self = CHECK(L, 1);
    luaL_checktype(L, 2, LUA_TTABLE);

    // Stack will be { top: string, table, self }
    lua_pushinteger(L, 1);
    lua_gettable(L, 2);
    const std::string_view objectName = CHECK_STRING_VIEW(L, -1);

    const lua_Unsigned len = lua_rawlen(L, 2);
    if (len == 2)
        return self->exportFunction(L, 2, objectName);

    else if (len == 3) {
        lua_pushinteger(L, 2);
        lua_gettable(L, 2);
        const std::string_view jobItTypeStr = CHECK_STRING_VIEW(L, -1);
        return self->exportJob(L, 2, objectName, jobItTypeStr);
    }

    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::exportFunction(lua_State *const L, const int tableIndex,
                                  const std::string_view name) noexcept
{
    auto *const context             = LuaContext::getContext(L);
    FunctionDeclaration *const func = context->getFunction(moduleName, name);
    out(context) << "Export function " << moduleName << "::" << name;

    lua_pushinteger(L, 2);
    lua_gettable(L, tableIndex);
    func->getLuaFunctionFromStack(L);

    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::exportJob(lua_State *const L, const int tableIndex, const std::string_view name,
                             const std::string_view itTypeStr) noexcept
{
    auto *const context             = LuaContext::getContext(L);
    const JobIteratorType jobItType = getJobIteratorTypeFromString(itTypeStr);
    JobDeclaration *const job       = context->getJob(moduleName, name);

    if (jobItType != JobIteratorType::Line && jobItType != JobIteratorType::Syllabe) {
        const std::string str(itTypeStr);
        luaL_error(L, "The iteration type must be LINE or SYLLABE, got %s", str.c_str());
    }

    else {
        out(context) << "Export job " << moduleName << "::" << name << "{" << itTypeStr << "}";

        lua_pushinteger(L, 3);
        lua_gettable(L, tableIndex);
        job->getLuaFunctionFromStack(L, jobItType);
    }

    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::setName(lua_State *const L) noexcept
{
    auto *const self = ModuleDeclaration::CHECK(L, 1);
    self->moduleName = CHECK_STRING_VIEW(L, 2);
    Utils::trim(self->moduleName);
    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::setAuthorName(lua_State *const L) noexcept
{
    auto *const self = ModuleDeclaration::CHECK(L, 1);
    self->authorName = CHECK_STRING_VIEW(L, 2);
    Utils::trim(self->authorName);
    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::setRevision(lua_State *const L) noexcept
{
    auto *const self = ModuleDeclaration::CHECK(L, 1);
    self->revision   = CHECK_STRING_VIEW(L, 2);
    Utils::trim(self->revision);
    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::setDescription(lua_State *const L) noexcept
{
    auto *const self        = ModuleDeclaration::CHECK(L, 1);
    self->moduleDescription = CHECK_STRING_VIEW(L, 2);
    Utils::trim(self->moduleDescription);
    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::setImports(lua_State *const L) noexcept
{
    auto *const self = ModuleDeclaration::CHECK(L, 1);
    IterateOverArray(L, 2, [self, L]() noexcept -> bool {
        std::string import(CHECK_STRING_VIEW(L, -1));
        Utils::trim(import);
        self->importNames.emplace_back(import);
        return false; // Value was not popped from stack!
    });
    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::setOptions(lua_State *const L) noexcept
{
    auto *const self = ModuleDeclaration::CHECK(L, 1);
    IterateOverArray(L, 2, [self, L]() noexcept -> bool {
        OptionDeclaration *const opt = OptionDeclaration::CHECK(L, -1);
        self->moduleOptions.emplace_back(opt);
        self->addUserValue(L, 1); // Tie the opt to the module
        return true;              // The value was popped from the stack
    });
    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::setFunctions(lua_State *const L) noexcept
{
    auto *const self = ModuleDeclaration::CHECK(L, 1);
    IterateOverArray(L, 2, [self, L]() noexcept -> bool {
        FunctionDeclaration *const fun = FunctionDeclaration::CHECK(L, -1);
        if (fun->isLocalToModule())
            fun->parentScript = self->moduleName;
        self->moduleFunctions.emplace_back(fun);
        self->addUserValue(L, 1); // Tie the function to the module
        return true;              // The value was popped from the stack
    });
    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::setJobs(lua_State *const L) noexcept
{
    auto *const self = ModuleDeclaration::CHECK(L, 1);
    IterateOverArray(L, 2, [self, L]() noexcept -> bool {
        JobDeclaration *const job = JobDeclaration::CHECK(L, -1);
        if (job->isLocalToModule())
            job->parentScript = self->moduleName;
        self->moduleJobs.emplace_back(job);
        self->addUserValue(L, 1); // Tie the job to the module
        return true;              // The value was popped from the stack
    });
    LUA_RETURN_NOTHING(L);
}

int
ModuleDeclaration::pushToRuntime(lua_State *const L) noexcept
{
    ModuleDeclaration *const self = CHECK(L, 1);
    auto *const context           = LuaContext::getContext(L);

    self->importNames.emplace_back(self->moduleName);   // You can use localy defined things
    Utils::uniqAndSort<std::string>(self->importNames); // Better for all std algorithms

    self->validateModule(L); // May abort the LUA context
    err(context) << "Register module " << VIVY_LOG_QUOTED(self->moduleName) << " in the runtime!";
    lua_settop(L, 1);

    if (context->registerDeclaration(self) == LuaContext::Code::Error)
        context->failedWith("Failed to register module " + self->moduleName);

    TODO(Import needed modules here !)

    LUA_RETURN_NOTHING(L);
}

void
ModuleDeclaration::validateModule(lua_State *const L) const noexcept
{
    // Minimal module file
    if (moduleName.empty() || authorName.empty())
        luaGlobalError(L, "The module does not have the minimal required information");

    // Imports
    {
        std::vector<std::string> usedImports{};
        const auto getUsedImports = [&usedImports](auto vectorOfPtr) noexcept -> void {
            for (const auto *ptr : vectorOfPtr) {
                if (!ptr->parentScript.empty())
                    usedImports.emplace_back(ptr->parentScript);
            }
        };

        getUsedImports(moduleJobs);
        getUsedImports(moduleFunctions);

        Utils::uniqAndSort<std::string>(usedImports);
        std::vector<std::string> unusedImports =
            Utils::sortedSetDifference(usedImports, importNames);

        if (!unusedImports.empty()) {
            std::size_t stringSize = 0;
            std::string listStrImports =
                "There are imported modules that are used withour begin declared:";
            for (const auto &str : unusedImports)
                stringSize += 1 + str.size();
            listStrImports.reserve(stringSize);
            for (const auto &str : unusedImports)
                listStrImports += " " + str;
            luaGlobalError(L, listStrImports);
        }
    }
}
