#include "PreCompiledHeaders.hh"
#include "Lib/Script/ScriptStore.hh"
#include "Lib/Uuid.hh"

using namespace Vivy;
using namespace std::string_literals;
using Code = Script::LuaContext::Code;

ScriptStore::ScriptStore() noexcept { resetLoadedScripts(); }

void
ScriptStore::loadScriptFolder(const QString &folderPath)
{
    QDir folder(folderPath);
    const auto filter = QDir::Files | QDir::NoSymLinks | QDir::Readable;

    // Load all module files, the loadDocument will ask for the implementation
    // file.
    for (const QString &file : folder.entryList(filter)) {
        if (!Utils::scriptFileSuffix.contains(QFileInfo(file).suffix()))
            continue;

        auto doc = loadDocument(file);
        if (luaContext->loadDocument(doc) != Code::Success) {
            const QString error = QStringLiteral("File: ") + file + QStringLiteral(" [") +
                                  doc->getUuid().toString() + QStringLiteral("]") +
                                  luaContext->getLastLuaError();
            throw std::runtime_error(error.toStdString());
        }
    }
}

std::vector<ScriptStore::Item>
ScriptStore::getLoadedScripts() const noexcept
{
    std::vector<ScriptStore::Item> ret;
    for (auto const &[uuid, docPtr] : documents)
        ret.emplace_back(Item{ uuid, docPtr->getName() });
    return ret;
}

void
ScriptStore::resetLoadedScripts() noexcept
{
    luaContext.reset(new Script::LuaContext());
}

bool
ScriptStore::executeScript(Uuid id) noexcept
{
    const auto &script = documents.at(id);
    return luaContext->loadDocument(script) == Code::Success;
}

bool
ScriptStore::executeScript(Uuid id, std::shared_ptr<Vivy::VivyDocument> doc) noexcept
{
    if (doc == nullptr)
        return false;
    if (luaContext->loadDocument(doc) != Script::LuaContext::Success) {
        luaContext->setFailedWith("A VivyDocument is already loaded, can't attach a new one");
        return false;
    }
    return executeScript(id);
}

const std::vector<std::string_view>
ScriptStore::getLoadedModules() const noexcept
{
    return luaContext ? luaContext->getAllModuleName() : std::vector<std::string_view>{};
}

const Script::ModuleDeclaration *
ScriptStore::getModule(const std::string_view str) const noexcept
{
    return luaContext->getModule(str);
}

QString
ScriptStore::getLastLuaError() noexcept
{
    return luaContext ? luaContext->getLastLuaError() : QString();
}
