#include "PreCompiledHeaders.hh"
#include "UI/ScriptViews/ScriptEditor.hh"
#include "UI/ScriptViews/ScriptHighlighter.hh"

using namespace Vivy;

ScriptEditor::LineNumberArea::LineNumberArea(ScriptEditor *editor) noexcept
    : QWidget(editor)
    , scriptEditor(editor)
{
}

QSize
ScriptEditor::LineNumberArea::sizeHint() const noexcept
{
    return QSize(scriptEditor->lineNumberAreaWidth(), 0);
}

void
ScriptEditor::LineNumberArea::paintEvent(QPaintEvent *event) noexcept
{
    scriptEditor->lineNumberAreaPaintEvent(event);
}

ScriptEditor::ScriptEditor(QWidget *parent) noexcept
    : QPlainTextEdit(parent)
{
    setStyleSheet(QStringLiteral("* {"
                                 "  background-color: #232629;"
                                 "  font-family: \"FiraCode\";"
                                 "  font-size: 10pt"
                                 "}"));
    lineNumberArea = new LineNumberArea(this);
    setFrameShape(QFrame::NoFrame);

    connect(this, &ScriptEditor::blockCountChanged, this, &ScriptEditor::updateLineNumberAreaWidth);
    connect(this, &ScriptEditor::updateRequest, this, &ScriptEditor::updateLineNumberArea);

    QTextCharFormat textFormat;
    textFormat.setForeground(QBrush(Qt::white));
    mergeCurrentCharFormat(textFormat);

    QPlainTextEdit::setCursorWidth(0);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setBackgroundVisible(true);
    updateLineNumberAreaWidth(0);
    setObjectName(QStringLiteral("Editor"));
    setFocus();
}

void
ScriptEditor::paintEvent(QPaintEvent *e) noexcept
{
    QPlainTextEdit::paintEvent(e);

    if (!cursorRect.isNull() && e->rect().intersects(cursorRect)) {
        QRect rect = cursorRect;
        cursorRect = QRect();
        QPlainTextEdit::viewport()->update(rect);
    }

    // Draw text cursor.
    QRect rect = QPlainTextEdit::cursorRect();
    if (e->rect().intersects(rect)) {
        QPainter painter(QPlainTextEdit::viewport());

        if (QPlainTextEdit::overwriteMode()) {
            QFontMetrics fm(QPlainTextEdit::font());
            const int position = QPlainTextEdit::textCursor().position();
            const QChar c      = QPlainTextEdit::document()->characterAt(position);
            rect.setWidth(fm.horizontalAdvance(c));
            painter.setPen(Qt::NoPen);
            painter.setBrush(QPlainTextEdit::palette().color(QPalette::Base));
            painter.setCompositionMode(QPainter::CompositionMode_Difference);
        } else {
            rect.setWidth(QPlainTextEdit::cursorWidth());
            painter.setPen(QPlainTextEdit::palette().color(QPalette::Text));
        }

        painter.drawRect(rect);
        cursorRect = rect;
    }
}

void
ScriptEditor::keyPressEvent(QKeyEvent *e) noexcept
{
    if (e->key() == Qt::Key_Tab) {
        QTextCursor cursor = textCursor();
        cursor.insertText(spacesForTab);
        e->accept();
    }

    else {
        QPlainTextEdit::keyPressEvent(e);
    }
}

int
ScriptEditor::lineNumberAreaWidth() noexcept
{
    int digits = 1;
    int max    = qMax(1, blockCount());
    while (max >= 10) {
        max /= 10;
        ++digits;
    }
    if (digits == 1)
        digits++;

    const int space = 3 + fontMetrics().horizontalAdvance('9') * digits;
    return space;
}

void
ScriptEditor::updateLineNumberAreaWidth([[maybe_unused]] int newBlockCount) noexcept
{
    setViewportMargins(lineNumberAreaWidth(), 0, 0, 0);
}

void
ScriptEditor::updateLineNumberArea(const QRect &rect, int dy) noexcept
{
    if (dy)
        lineNumberArea->scroll(0, dy);
    else
        lineNumberArea->update(0, rect.y(), lineNumberArea->width(), rect.height());

    if (rect.contains(viewport()->rect()))
        updateLineNumberAreaWidth(0);
}

void
ScriptEditor::resizeEvent(QResizeEvent *e) noexcept
{
    QPlainTextEdit::resizeEvent(e);
    const QRect cr = contentsRect();
    lineNumberArea->setGeometry(QRect(cr.left(), cr.top(), lineNumberAreaWidth(), cr.height()));
}

void
ScriptEditor::lineNumberAreaPaintEvent(QPaintEvent *event) noexcept
{
    QPainter painter(lineNumberArea);
    painter.fillRect(event->rect(), QColor::fromRgb(49, 54, 59));
    QTextBlock block  = firstVisibleBlock();
    int blockNumber   = block.blockNumber();
    int top           = qRound(blockBoundingGeometry(block).translated(contentOffset()).top());
    int bottom        = top + qRound(blockBoundingRect(block).height());
    const int current = textCursor().blockNumber();

    while (block.isValid() && top <= event->rect().bottom()) {
        if (block.isVisible() && bottom >= event->rect().top()) {
            const bool drawCurrent = blockNumber == current;
            const QString number =
                drawCurrent ? QStringLiteral("->") : QString::number(blockNumber + 1);
            painter.setPen(drawCurrent ? Qt::white : Qt::black);
            painter.drawText(0, top, lineNumberArea->width(), fontMetrics().height(),
                             Qt::AlignRight, number);
        }

        block  = block.next();
        top    = bottom;
        bottom = top + qRound(blockBoundingRect(block).height());
        ++blockNumber;
    }
}

void
ScriptEditor::updateLastLuaError(int line, QString desc)
{
    logDebug() << "Update error on line " << line << " with description " << desc;
}
