#include "PreCompiledHeaders.hh"
#include "UI/ScriptViews/ScriptHighlighter.hh"
#include "UI/Theme/Theme.hh"
#include "VivyApplication.hh"
#include "Lib/Utils.hh"

using namespace Vivy;

ScriptHighlighter::HighlightingRule::HighlightingRule(QRegularExpression pttrn, QTextCharFormat frmt)
    : pattern(pttrn)
    , format(frmt)
{
}

void
ScriptHighlighter::resetHighlightingRule() noexcept
{
    highlightingRules.clear();
    const Theme::HighlightingTheme &theme = vivyApp->getTheme()->getHighlightingTheme();

    // function calls
    functionFormat.setForeground(theme.functionForeground);
    highlightingRules.emplace_back(QRegularExpressionLiteral("\\b[A-Za-z0-9_]+[ ]*(?=\\()"), functionFormat);
    highlightingRules.emplace_back(QRegularExpressionLiteral("\\b[A-Za-z0-9_]+[ ]*(?=\\{)"), functionFormat);

    // Member and enum members
    enumFormat.setForeground(theme.enumForeground);
    enumFormat.setFontWeight(QFont::Bold);
    memberFormat.setForeground(theme.memberForeground);
    memberFormat.setFontWeight(QFont::StyleItalic);
    highlightingRules.emplace_back(QRegularExpressionLiteral("\\.[a-z][\\dA-Za-z]*\\b"), memberFormat);
    highlightingRules.emplace_back(QRegularExpressionLiteral("\\b[A-Z][\\dA-Z_]*\\b"), enumFormat);

    // keywords
    const QStringList keywordPatterns = {
        QStringLiteral("\\bfunction\\b"), QStringLiteral("\\bbreak\\b"),
        QStringLiteral("\\bgoto\\b"),     QStringLiteral("\\bdo\\b"),
        QStringLiteral("\\bend\\b"),      QStringLiteral("\\bwhile\\b"),
        QStringLiteral("\\brepeat\\b"),   QStringLiteral("\\buntil\\b"),
        QStringLiteral("\\bif\\b"),       QStringLiteral("\\bthen\\b"),
        QStringLiteral("\\belseif\\b"),   QStringLiteral("\\belse\\b"),
        QStringLiteral("\\bfor\\b"),      QStringLiteral("\\bin\\b"),
        QStringLiteral("\\blocal\\b"),    QStringLiteral("\\bor\\b"),
        QStringLiteral("\\band\\b"),      QStringLiteral("\\bnot\\b"),
        QStringLiteral("\\breturn\\b"),   QStringLiteral("\\b\\=\\b"),
        QStringLiteral("\\b\\+\\=\\b"),   QStringLiteral("\\b\\-\\=\\b"),
        QStringLiteral("\\b\\*\\=\\b"),   QStringLiteral("\\b\\/\\=\\b"),
    };

    keywordFormat.setForeground(theme.keywordForeground);
    keywordFormat.setFontWeight(QFont::Bold);

    for (QString const &pattern : keywordPatterns)
        highlightingRules.emplace_back(QRegularExpression(pattern), keywordFormat);

    // numbers, boolean, nil
    const QStringList valuePatterns = {
        QStringLiteral("\\bnil\\b"),
        QStringLiteral("\\btrue\\b"),
        QStringLiteral("\\bfalse\\b"),
        QStringLiteral("\\b\\d+\\b"),
        QStringLiteral("\\b\\d+.\\b"),
        QStringLiteral("\\b\\d+e\\b"),
        QStringLiteral("\\b\\[\\dA-Fa-F]+\\b"),
    };

    valueFormat.setForeground(theme.valueForeground);
    valueFormat.setFontWeight(QFont::Normal);

    for (QString const &pattern : valuePatterns)
        highlightingRules.emplace_back(QRegularExpression(pattern), valueFormat);

    // strings
    quotationFormat.setForeground(theme.valueForeground);
    highlightingRules.emplace_back(QRegularExpressionLiteral("\"[^\"]*\""), quotationFormat);
    highlightingRules.emplace_back(QRegularExpressionLiteral("\'[^\']*\'"), quotationFormat);
    quoteStartExpression = QRegularExpressionLiteral("\\[\\[");
    quoteEndExpression   = QRegularExpressionLiteral("\\]\\]");

    // comments
    singleLineCommentFormat.setForeground(theme.commentForeground);
    highlightingRules.emplace_back(QRegularExpressionLiteral("--[^\n]*"), singleLineCommentFormat);
    commentStartExpression = QRegularExpressionLiteral("--\\[\\[");
    commentEndExpression   = QRegularExpressionLiteral("\\]\\]");
}

ScriptHighlighter::ScriptHighlighter(QTextDocument *parent) noexcept
    : QSyntaxHighlighter(parent)
{
    resetHighlightingRule();
}

void
ScriptHighlighter::setCurrentBlockState(const ScriptHighlighter::BlockState state) noexcept
{
    QSyntaxHighlighter::setCurrentBlockState(Utils::toUnderlying(state));
}

ScriptHighlighter::BlockState
ScriptHighlighter::previousBlockState() noexcept
{
    const int state = QSyntaxHighlighter::previousBlockState();
    if (state != Utils::toUnderlying(BlockState::Quote) &&
        state != Utils::toUnderlying(BlockState::Comment))
        return BlockState::None;

    return static_cast<BlockState>(state);
}

void
ScriptHighlighter::highlightBlock(const QString &text) noexcept
{
    for (const HighlightingRule &rule : highlightingRules) {
        QRegularExpression expression(rule.pattern);
        QRegularExpressionMatch matchStart = expression.match(text);
        int index = int(matchStart.capturedStart(0));
        while (index >= 0) {
            const int length = int(matchStart.capturedLength(0));
            setFormat(index, length, rule.format);
            // FIXME: unused, can delete?
            QRegularExpressionMatch matchEnd = expression.match(text, index + length);
            index = int(matchEnd.capturedStart(0));
        }
    }

    setCurrentBlockState(BlockState::None);
    const BlockState previousState = previousBlockState();

    highlightStringBlock(text, previousState);
    highlightCommentBlock(text, previousState);
}

void
ScriptHighlighter::highlightBlock(const QString &text, const BlockState previousState,
                                  const BlockState toHighlight, const QRegularExpression &startExpr,
                                  const QRegularExpression &endExpr, const QTextCharFormat &format) noexcept
{
    int start = -1;
    if (previousState == toHighlight) {
        start = 0;
    } else if (previousState == BlockState::None) {
        QRegularExpressionMatch match = startExpr.match(text);
        start = int(startExpr.match(text).capturedStart(0));
    }

    while (start >= 0) {
        const int end = int(endExpr.match(text, start).capturedStart(0));
        int length;

        if (end == -1) {
            setCurrentBlockState(toHighlight);
            length = int(text.length()) - start;
        } else {
            length = end - start + int(endExpr.match(text).capturedLength(0));
        }

        setFormat(start, length, format);
        // FIXME: unused, can delete? or expected as startExpr reference?
        start = int(startExpr.match(text, start + length).capturedStart(0));
    }
}

void
ScriptHighlighter::highlightStringBlock(const QString &text,
                                        const BlockState previousState) noexcept
{
    highlightBlock(text, previousState, BlockState::Quote, quoteStartExpression, quoteEndExpression,
                   quotationFormat);
}

void
ScriptHighlighter::highlightCommentBlock(const QString &text,
                                         const BlockState previousState) noexcept
{
    highlightBlock(text, previousState, BlockState::Comment, commentStartExpression,
                   commentEndExpression, singleLineCommentFormat);
}
