#include "PreCompiledHeaders.hh"
#include "UI/Theme/Theme.hh"
#include "VivyApplication.hh"
#include "Lib/HostOsInfo.hh"
#if VIVY_MACOS
#import "ThemeMac.hh"
#endif

namespace Vivy
{
struct Theme::Private {
    Private() noexcept;

    ThemeType classType{ ThemeType::System };
    VivyThemeType preferedColor{ VivyThemeType::Dark };

    QString id;
    QString fileName;
    QString displayName;
    QStringList preferredStyles;
    QString defaultTextEditorColorScheme;
    QVector<QPair<QColor, QString>> colors;
    QVector<QString> imageFiles;
    QVector<QGradientStops> gradients;
    QVector<bool> flags;
    QMap<QString, QColor> palette;
};

// If you copy QPalette, default values stay at default, even if that default is different
// within the context of different widgets. Create deep copy.
static QPalette
copyPalette(const QPalette &p) noexcept
{
    QPalette res;
    for (int group = 0; group < QPalette::NColorGroups; ++group) {
        for (int role = 0; role < QPalette::NColorRoles; ++role) {
            res.setBrush(QPalette::ColorGroup(group), QPalette::ColorRole(role),
                         p.brush(QPalette::ColorGroup(group), QPalette::ColorRole(role)));
        }
    }
    return res;
}

[[maybe_unused]] static bool
systemUsesDarkMode() noexcept
{
    if (Utils::HostOsInfo::isWindowsHost()) {
        constexpr char regkey[] =
            "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Themes\\Personalize";
        bool ok;
        const auto setting =
            QSettings(regkey, QSettings::NativeFormat).value("AppsUseLightTheme").toInt(&ok);
        return ok && setting == 0;
    }

    else if constexpr (Utils::HostOsInfo::isMacHost())
        return false;

    else if constexpr (Utils::HostOsInfo::isLinuxHost())
        return true;

    return false;
}

static QPalette
initialPalette() noexcept
{
    static QPalette palette = copyPalette(QApplication::palette());
    return palette;
}

Theme::Private::Private() noexcept
{
    const QMetaObject &m = Theme::staticMetaObject;
    colors.resize(m.enumerator(m.indexOfEnumerator("Color")).keyCount());
    imageFiles.resize(m.enumerator(m.indexOfEnumerator("ImageFile")).keyCount());
    gradients.resize(m.enumerator(m.indexOfEnumerator("Gradient")).keyCount());
    flags.resize(m.enumerator(m.indexOfEnumerator("Flag")).keyCount());
}

void
Theme::applyToApplication() const noexcept
{
    if (d->classType == ThemeType::QssFile) {
        QFile stylesheet(d->fileName);
        if (!stylesheet.exists()) {
            qFatal("Missing stylesheet");
        } else {
            stylesheet.open(QFile::ReadOnly | QFile::Text);
            QTextStream stylesheetStream(&stylesheet);
            vivyApp->setStyleSheet(stylesheetStream.readAll());
        }
    }

    else {
#if VIVY_MACOS
        // Match the native UI theme and palette with the creator
        // theme by forcing light aqua for light creator themes.
        if (!flag(Theme::DarkUserInterface))
            Utils::forceMacOSLightAquaApperance();
#endif
        QApplication::setPalette(palette());
    }
}

Theme::Theme(const QString &id, QObject *parent) noexcept
    : QObject(parent)
    , d(new Private)
{
    d->id = id;
}

Theme::~Theme() noexcept { delete d; }

QStringList
Theme::preferredStyles() const noexcept
{
    return d->preferredStyles;
}

QString
Theme::defaultTextEditorColorScheme() const noexcept
{
    return d->defaultTextEditorColorScheme;
}

QString
Theme::id() const noexcept
{
    return d->id;
}

bool
Theme::flag(Theme::Flag f) const noexcept
{
    return d->flags[f];
}

QColor
Theme::color(Theme::Color role) const noexcept
{
    return d->colors[role].first;
}

QString
Theme::imageFile(Theme::ImageFile imageFile, const QString &fallBack) const noexcept
{
    const QString &file = d->imageFiles.at(imageFile);
    return file.isEmpty() ? fallBack : file;
}

QGradientStops
Theme::gradient(Theme::Gradient role) const noexcept
{
    return d->gradients[role];
}

QPair<QColor, QString>
Theme::readNamedColor(const QString &color) const
{
    if (d->palette.contains(color))
        return qMakePair(d->palette[color], color);
    if (color == QLatin1String("style"))
        return qMakePair(QColor(), QString());

    const QColor col('#' + color);
    if (!col.isValid()) {
        logWarning() << "Color " << VIVY_LOG_QUOTED(qPrintable(color))
                     << " is neither a named color nor a valid color";
        return qMakePair(Qt::black, QString());
    }
    return qMakePair(col, QString());
}

QString
Theme::filePath() const noexcept
{
    return d->fileName;
}

QString
Theme::displayName() const noexcept
{
    return d->displayName;
}

void
Theme::setDisplayName(const QString &name) noexcept
{
    d->displayName = name;
}

Theme *
Theme::fromQssFile(const QString &id, const QString &file) noexcept
{
    Theme *ret        = new Theme(id);
    ret->d->fileName  = file;
    ret->d->classType = ThemeType::QssFile;
    ret->setDisplayName(QFileInfo(file).baseName());
    return ret;
}

Theme *
Theme::fromSettings(const QString &id, QSettings *const settings) noexcept
{
    Theme *ret = new Theme(id);
    ret->readSettings(settings);
    return ret;
}

Theme *
Theme::fromVivyTheme(const QString &id, const VivyThemeType which) noexcept
{
    Theme *ret = fromQssFile(id, which == VivyThemeType::Dark
                                     ? QStringLiteral(":qdarkstyle/dark/style.qss")
                                     : QStringLiteral(":qdarkstyle/light/style.qss"));
    ret->setDisplayName(QStringLiteral("Vivy Theme"));
    return ret;
}

const Theme::HighlightingTheme &
Theme::getHighlightingTheme() const noexcept
{
    return highlightingTheme[Utils::toUnderlying(d->preferedColor)];
}

void
Theme::readSettings(QSettings *const settings) noexcept
{
    d->fileName          = settings->fileName();
    d->classType         = ThemeType::QtCreator;
    const QMetaObject &m = *metaObject();

    {
        d->displayName =
            settings->value(QLatin1String("ThemeName"), QLatin1String("unnamed")).toString();
        d->preferredStyles = settings->value(QLatin1String("PreferredStyles")).toStringList();
        d->preferredStyles.removeAll(QString());
        d->defaultTextEditorColorScheme =
            settings->value(QLatin1String("DefaultTextEditorColorScheme")).toString();
    }
    {
        settings->beginGroup(QLatin1String("Palette"));
        const QStringList allKeys = settings->allKeys();
        for (const QString &key : allKeys)
            d->palette[key] = readNamedColor(settings->value(key).toString()).first;
        settings->endGroup();
    }
    {
        settings->beginGroup(QLatin1String("Colors"));
        QMetaEnum e = m.enumerator(m.indexOfEnumerator("Color"));
        for (int i = 0, total = e.keyCount(); i < total; ++i) {
            const QString key = QLatin1String(e.key(i));
            if (!settings->contains(key)) {
                if (i < PaletteWindow || i > PalettePlaceholderTextDisabled)
                    logWarning()
                        << "Theme " << VIVY_LOG_QUOTED(qPrintable(d->fileName))
                        << " misses color setting for key " << VIVY_LOG_QUOTED(qPrintable(key));
                continue;
            }
            d->colors[i] = readNamedColor(settings->value(key).toString());
        }
        settings->endGroup();
    }
    {
        settings->beginGroup(QLatin1String("ImageFiles"));
        QMetaEnum e = m.enumerator(m.indexOfEnumerator("ImageFile"));
        for (int i = 0, total = e.keyCount(); i < total; ++i) {
            const QString key = QLatin1String(e.key(i));
            d->imageFiles[i]  = settings->value(key).toString();
        }
        settings->endGroup();
    }
    {
        settings->beginGroup(QLatin1String("Gradients"));
        QMetaEnum e = m.enumerator(m.indexOfEnumerator("Gradient"));
        for (int i = 0, total = e.keyCount(); i < total; ++i) {
            const QString key = QLatin1String(e.key(i));
            QGradientStops stops;
            int size = settings->beginReadArray(key);
            for (int j = 0; j < size; ++j) {
                settings->setArrayIndex(j);
                VIVY_ASSERT(settings->contains(QLatin1String("pos")));
                VIVY_ASSERT(settings->contains(QLatin1String("color")));
                const double pos = settings->value(QLatin1String("pos")).toDouble();
                const QColor c('#' + settings->value(QLatin1String("color")).toString());
                stops.append(qMakePair(pos, c));
            }
            settings->endArray();
            d->gradients[i] = stops;
        }
        settings->endGroup();
    }
    {
        settings->beginGroup(QLatin1String("Flags"));
        QMetaEnum e = m.enumerator(m.indexOfEnumerator("Flag"));
        for (int i = 0, total = e.keyCount(); i < total; ++i) {
            const QString key = QLatin1String(e.key(i));
            VIVY_ASSERT(settings->contains(key));
            d->flags[i] = settings->value(key).toBool();
        }
        settings->endGroup();
    }
}

QPalette
Theme::palette() const noexcept
{
    QPalette pal = initialPalette();

    const static struct {
        Color themeColor;
        QPalette::ColorRole paletteColorRole;
        QPalette::ColorGroup paletteColorGroup;
        bool setColorRoleAsBrush;
    } mapping[] = {
        { PaletteWindow, QPalette::Window, QPalette::All, false },
        { PaletteWindowDisabled, QPalette::Window, QPalette::Disabled, false },
        { PaletteWindowText, QPalette::WindowText, QPalette::All, true },
        { PaletteWindowTextDisabled, QPalette::WindowText, QPalette::Disabled, true },
        { PaletteBase, QPalette::Base, QPalette::All, false },
        { PaletteBaseDisabled, QPalette::Base, QPalette::Disabled, false },
        { PaletteAlternateBase, QPalette::AlternateBase, QPalette::All, false },
        { PaletteAlternateBaseDisabled, QPalette::AlternateBase, QPalette::Disabled, false },
        { PaletteToolTipBase, QPalette::ToolTipBase, QPalette::All, true },
        { PaletteToolTipBaseDisabled, QPalette::ToolTipBase, QPalette::Disabled, true },
        { PaletteToolTipText, QPalette::ToolTipText, QPalette::All, false },
        { PaletteToolTipTextDisabled, QPalette::ToolTipText, QPalette::Disabled, false },
        { PaletteText, QPalette::Text, QPalette::All, true },
        { PaletteTextDisabled, QPalette::Text, QPalette::Disabled, true },
        { PaletteButton, QPalette::Button, QPalette::All, false },
        { PaletteButtonDisabled, QPalette::Button, QPalette::Disabled, false },
        { PaletteButtonText, QPalette::ButtonText, QPalette::All, true },
        { PaletteButtonTextDisabled, QPalette::ButtonText, QPalette::Disabled, true },
        { PaletteBrightText, QPalette::BrightText, QPalette::All, false },
        { PaletteBrightTextDisabled, QPalette::BrightText, QPalette::Disabled, false },
        { PaletteHighlight, QPalette::Highlight, QPalette::All, true },
        { PaletteHighlightDisabled, QPalette::Highlight, QPalette::Disabled, true },
        { PaletteHighlightedText, QPalette::HighlightedText, QPalette::All, true },
        { PaletteHighlightedTextDisabled, QPalette::HighlightedText, QPalette::Disabled, true },
        { PaletteLink, QPalette::Link, QPalette::All, false },
        { PaletteLinkDisabled, QPalette::Link, QPalette::Disabled, false },
        { PaletteLinkVisited, QPalette::LinkVisited, QPalette::All, false },
        { PaletteLinkVisitedDisabled, QPalette::LinkVisited, QPalette::Disabled, false },
        { PaletteLight, QPalette::Light, QPalette::All, false },
        { PaletteLightDisabled, QPalette::Light, QPalette::Disabled, false },
        { PaletteMidlight, QPalette::Midlight, QPalette::All, false },
        { PaletteMidlightDisabled, QPalette::Midlight, QPalette::Disabled, false },
        { PaletteDark, QPalette::Dark, QPalette::All, false },
        { PaletteDarkDisabled, QPalette::Dark, QPalette::Disabled, false },
        { PaletteMid, QPalette::Mid, QPalette::All, false },
        { PaletteMidDisabled, QPalette::Mid, QPalette::Disabled, false },
        { PaletteShadow, QPalette::Shadow, QPalette::All, false },
        { PaletteShadowDisabled, QPalette::Shadow, QPalette::Disabled, false },
        { PalettePlaceholderText, QPalette::PlaceholderText, QPalette::All, false },
        { PalettePlaceholderTextDisabled, QPalette::PlaceholderText, QPalette::Disabled, false },
    };

    for (auto entry : mapping) {
        const QColor themeColor = color(entry.themeColor);
        // Use original color if color is not defined in theme.
        if (themeColor.isValid()) {
            if (entry.setColorRoleAsBrush)
                // TODO: Find out why sometimes setBrush is used
                pal.setBrush(entry.paletteColorGroup, entry.paletteColorRole, themeColor);
            else
                pal.setColor(entry.paletteColorGroup, entry.paletteColorRole, themeColor);
        }
    }

    return pal;
}
}
